@@ -1,6 +1,6 @@
 /* tmul_2si -- test file for mpc_mul_2si.
 
-Copyright (C) 2012 INRIA
+Copyright (C) 2012, 2013 INRIA
 
 This file is part of GNU MPC.
 
@@ -20,14 +20,19 @@ along with this program. If not, see http://www.gnu.org/licenses/ .
 
 #include "mpc-tests.h"
 
+#define MPC_FUNCTION_CALL                                               \
+  P[0].mpc_inex = mpc_mul_2si (P[1].mpc, P[2].mpc, P[3].si, P[4].mpc_rnd)
+#define MPC_FUNCTION_CALL_REUSE_OP1                                     \
+  P[0].mpc_inex = mpc_mul_2si (P[1].mpc, P[1].mpc, P[3].si, P[4].mpc_rnd)
+
+#include "tgeneric.tpl"
+
 int
 main (void)
 {
-  DECL_FUNC (CCS, f, mpc_mul_2si);
-
   test_start ();
 
-  tgeneric (f, 2, 1024, 7, -1);
+  tgeneric_template ("mul_2si.dsc", 2, 1024, 7, 1024);
 
   test_end ();
 
